# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import warnings
from typing import List, Optional

import torch

from torchlie.functional import SO3
from torchlie.functional.constants import DeviceType


class Link(abc.ABC):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_joint: "Joint" = None,
        child_joints: Optional[List["Joint"]] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        self._name = name
        self._id = id
        # self._parent_joint is None means that the link is rigidly connected to the world
        self._parent_joint = parent_joint
        self._child_joints = child_joints if child_joints else []
        self._ancestor_links: List[Link] = []
        self._ancestor_non_fixed_joint_ids: List[int] = []
        self._dtype = dtype

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> int:
        return self._id

    @property
    def parent_link(self) -> "Link":
        return (
            self._parent_joint.parent_link if self._parent_joint is not None else None
        )

    @property
    def parent_joint(self) -> "Joint":
        return self._parent_joint

    @property
    def child_joints(self) -> List["Joint"]:
        return self._child_joints

    @property
    def ancestor_links(self) -> List["Link"]:
        return self._ancestor_links

    @property
    def ancestor_non_fixed_joint_ids(self):
        return self._ancestor_non_fixed_joint_ids

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    def _update_ancestor_links(self):
        curr = self.parent_link
        self._ancestor_links = []
        while curr is not None:
            self._ancestor_links.insert(0, curr)
            curr = curr.parent_link

    def _add_child_joint(self, child_joint: "Joint"):
        self._child_joints.append(child_joint)

    def _remove_child_joint(self, child_joint: "Joint"):
        self._child_joints.remove(child_joint)


class Joint(abc.ABC):
    def __init__(
        self,
        name: str,
        dof: int,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        if origin is None and dtype is None:
            dtype = torch.get_default_dtype()
        if origin is None and device is None:
            device = torch.device("cpu")
        if origin is not None:
            self._origin = origin

            if dtype is not None and origin.dtype != dtype:
                warnings.warn(
                    f"The origin's dtype {origin.dtype} does not match given dtype {dtype}, "
                    "Origin's dtype will take precendence."
                )
            dtype = origin.dtype

            if device is not None and origin.device != device:
                warnings.warn(
                    f"tensor.device {origin.device} does not match given device {device}, "
                    "tensor.device will take precendence."
                )
            dtype = origin.dtype
            device = origin.device
        else:
            origin = torch.zeros(1, 3, 4, dtype=dtype, device=device)
            origin[:, 0, 0] = 1
            origin[:, 1, 1] = 1
            origin[:, 2, 2] = 1

        self._name = name
        self._dof = dof
        self._id = id
        self._parent_link = parent_link
        self._child_link = child_link
        self._origin = origin
        self._dtype = dtype
        self._device = torch.device(device)
        self._axis: torch.Tensor = torch.zeros(6, self.dof, dtype=dtype, device=device)

    @property
    def name(self) -> str:
        return self._name

    @property
    def dof(self) -> int:
        return self._dof

    @property
    def id(self) -> int:
        return self._id

    @property
    def parent_link(self) -> Optional[Link]:
        return self._parent_link

    @property
    def child_link(self) -> Optional[Link]:
        return self._child_link

    @property
    def origin(self) -> torch.Tensor:
        return self._origin

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def axis(self) -> torch.Tensor:
        return self._axis

    @abc.abstractmethod
    def relative_pose(self, angle: Optional[torch.Tensor] = None) -> torch.Tensor:
        pass


class FixedJoint(Joint):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, 0, id, parent_link, child_link, origin, dtype, device)

    def relative_pose(self, *args) -> torch.Tensor:
        if len(args) != 0:
            raise ValueError("Fixed joint has no inputs.")
        return self.origin


class _RevoluteJointImpl(Joint):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, 1, id, parent_link, child_link, origin, dtype, device)

    @abc.abstractmethod
    def _rotation_impl(self, angle: torch.Tensor) -> torch.Tensor:
        pass

    @staticmethod
    def _check_input(angle: torch.Tensor):
        if angle.ndim == 1:
            angle = angle.view(-1, 1)
        if angle.ndim != 2 or angle.shape[1] != 1:
            raise ValueError("The joint angle must be a vector.")

    def rotation(self, angle: torch.Tensor) -> torch.Tensor:
        _RevoluteJointImpl._check_input(angle)
        return self._rotation_impl(angle.view(-1, 1))

    def _relative_pose_impl(self, angle: torch.Tensor) -> torch.Tensor:
        rot = self.rotation(angle)
        ret = angle.new_empty(angle.shape[0], 3, 4)
        ret[:, :, :3] = self.origin[:, :, :3] @ rot
        ret[:, :, 3] = self.origin[:, :, 3]
        return ret

    def relative_pose(self, *args) -> torch.Tensor:
        if len(args) != 1:
            raise ValueError("Revolute joint requires one input.")
        angle: torch.Tensor = args[0]
        return self._relative_pose_impl(angle)


class RevoluteJoint(_RevoluteJointImpl):
    def __init__(
        self,
        name: str,
        revolute_axis: torch.Tensor,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        if revolute_axis.numel() != 3:
            raise ValueError("The revolute axis must have 3 elements.")

        super().__init__(name, id, parent_link, child_link, origin, dtype, device)

        if revolute_axis.dtype != self.dtype:
            raise ValueError(f"The dtype of revolute_axis should be {self.dtype}.")

        if revolute_axis.device != self.device:
            raise ValueError(f"The device of revolute_axis should be {self.device}.")

        self._axis[3:] = revolute_axis.view(-1, 1)

    def _rotation_impl(self, angle: torch.Tensor) -> torch.Tensor:
        return SO3.exp(angle @ self.axis[3:].view(1, -1))


class _RevoluteJointXYZImpl(_RevoluteJointImpl):
    axis_info = [[0, 1, 2], [1, 2, 0], [2, 0, 1]]

    def __init__(
        self,
        name: str,
        axis_id: int,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, parent_link, child_link, origin, dtype, device)
        if axis_id < 0 or axis_id >= 3:
            raise ValueError("The axis_id must be one of (0, 1, 2).")
        self._axis_id = axis_id
        self._axis[self.axis_id + 3] = 1

    @property
    def axis_id(self) -> int:
        return self._axis_id

    def _rotation_impl(self, angle: torch.Tensor) -> torch.Tensor:
        info = _RevoluteJointXYZImpl.axis_info[self.axis_id]
        rot = angle.new_zeros(angle.shape[0], 3, 3)
        rot[:, info[0], info[0]] = 1
        rot[:, info[1], info[1]] = angle.cos()
        rot[:, info[2], info[2]] = rot[:, info[1], info[1]]
        rot[:, info[2], info[1]] = angle.sin()
        rot[:, info[1], info[2]] = -rot[:, info[2], info[1]]

        return rot


class RevoluteJointX(_RevoluteJointXYZImpl):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, 0, parent_link, child_link, origin, dtype, device)


class RevoluteJointY(_RevoluteJointXYZImpl):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, 1, parent_link, child_link, origin, dtype, device)


class RevoluteJointZ(_RevoluteJointXYZImpl):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, 2, parent_link, child_link, origin, dtype, device)


class _PrismaticJointImpl(Joint):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, 1, id, parent_link, child_link, origin, dtype, device)

    @abc.abstractmethod
    def _translation_impl(self, angle: torch.Tensor) -> torch.Tensor:
        pass

    @staticmethod
    def _check_input(angle: torch.Tensor):
        if angle.ndim == 1:
            angle = angle.view(-1, 1)
        if angle.ndim != 2 or angle.shape[1] != 1:
            raise ValueError("The joint angle must be a vector.")

    def translation(self, angle: torch.Tensor) -> torch.Tensor:
        _PrismaticJointImpl._check_input(angle)
        return self.translation(angle.view(-1, 1))

    def _relative_pose_impl(self, angle: torch.Tensor) -> torch.Tensor:
        trans = self.translation(angle)
        ret = angle.new_empty(angle.shape[0], 3, 4)
        ret[:, :, :3] = self.origin[:, :, :3]
        ret[:, :, 3:] = (
            self.origin[:, :, :3] @ trans.view(-1, 3, 1) + self.origin[:, :, 3:]
        )
        return ret

    def relative_pose(self, *args) -> torch.Tensor:
        if len(args) != 1:
            raise ValueError("Prismatic joint requires one input.")
        angle: torch.Tensor = args[0]
        return self._relative_pose_impl(angle)


class PrismaticJoint(_PrismaticJointImpl):
    def __init__(
        self,
        name: str,
        prismatic_axis: torch.Tensor,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        if prismatic_axis.numel() != 3:
            raise ValueError("The prismatic axis must have 3 elements.")

        super().__init__(name, id, parent_link, child_link, origin, dtype, device)

        if prismatic_axis.dtype != self.dtype:
            raise ValueError(f"The dtype of prismatic_axis should be {self.dtype}.")

        if prismatic_axis.device != self.device:
            raise ValueError(f"The device of prismatic_axis should be {self.device}.")

        self._axis[:3] = prismatic_axis.view(-1, 1)

    def _translation_impl(self, angle: torch.Tensor) -> torch.Tensor:
        return angle @ self.axis[:3].view(1, -1)


class _PrismaticJointXYZImpl(_PrismaticJointImpl):
    def __init__(
        self,
        name: str,
        axis_id: int,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, parent_link, child_link, origin, dtype, device)
        if axis_id < 0 or axis_id >= 3:
            raise ValueError("The axis_id must be one of (0, 1, 2).")
        self._axis_id = axis_id
        self._axis[axis_id] = 1

    @property
    def axis_id(self) -> int:
        return self._axis_id

    def _translation_impl(self, angle: torch.Tensor) -> torch.Tensor:
        trans = angle.new_zeros(angle.shape[0], 3)
        trans[:, self.axis_id] = angle
        return trans

    def _relative_pose_impl(self, angle: torch.Tensor) -> torch.Tensor:
        _PrismaticJointXYZImpl._check_input(angle)
        angle = angle.view(-1, 1)
        ret = angle.new_empty(angle.shape[0], 3, 4)
        ret[:, :, :3] = self.origin[:, :, :3]
        ret[:, :, 3] = angle * self.origin[:, :, self.axis_id] + self.origin[:, :, 3]
        return ret


class PrismaticJointX(_PrismaticJointXYZImpl):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, 0, parent_link, child_link, origin, dtype, device)


class PrismaticJointY(_PrismaticJointXYZImpl):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, 1, parent_link, child_link, origin, dtype, device)


class PrismaticJointZ(_PrismaticJointXYZImpl):
    def __init__(
        self,
        name: str,
        id: Optional[int] = None,
        parent_link: Optional[Link] = None,
        child_link: Optional[Link] = None,
        origin: Optional[torch.Tensor] = None,
        dtype: Optional[torch.dtype] = None,
        device: DeviceType = None,
    ):
        super().__init__(name, id, 2, parent_link, child_link, origin, dtype, device)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
__version__ = "0.1.0"

from .forward_kinematics import get_forward_kinematics_fns
from .robot import Robot



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional

import torch

from torchlie.functional import SE3

from .robot import Joint, Link, Robot

# TODO: Add support for joints with DOF>1


def ForwardKinematicsFactory(robot: Robot, link_names: Optional[List[str]] = None):
    links: List[Link] = robot.get_links(link_names)
    link_ids: List[int] = [link.id for link in links]

    ancestor_links: List[Link] = []
    joint_ids: List[int] = []
    for link in links:
        ancestor_links += [anc for anc in link.ancestor_links]
        joint_ids += link.ancestor_non_fixed_joint_ids
    related_link_ids = sorted(list(set([anc.id for anc in ancestor_links] + link_ids)))
    joint_ids = sorted(list(set(joint_ids)))

    def _forward_kinematics_helper(angles: torch.Tensor):
        if angles.ndim != 2 or angles.shape[1] != robot.dof:
            raise ValueError(
                f"Joint angles for {robot.name} should be {robot.dof}-D vectors"
            )

        poses: List[Optional[torch.Tensor]] = [None] * robot.num_links
        poses[0] = angles.new_zeros(angles.shape[0], 3, 4)
        poses[0][:, 0, 0] = 1
        poses[0][:, 1, 1] = 1
        poses[0][:, 2, 2] = 1

        for id in related_link_ids[1:]:
            curr: Link = robot.links[id]
            joint: Joint = robot.links[id].parent_joint
            prev: Link = joint.parent_link
            relative_pose = (
                joint.relative_pose(angles[:, joint.id])
                if joint.id < robot.dof
                else joint.relative_pose()
            )
            poses[curr.id] = SE3.compose(poses[prev.id], relative_pose)

        return poses

    def _forward_kinematics_impl(angles: torch.Tensor):
        poses = _forward_kinematics_helper(angles)
        return tuple(poses[id] for id in link_ids)

    def _forward_kinematics_backward_helper(
        poses: List[Optional[torch.Tensor]],
    ) -> torch.Tensor:
        jposes = poses[0].new_zeros(poses[0].shape[0], 6, robot.dof)

        for id in related_link_ids[1:]:
            link: Link = robot.links[id]
            joint: Joint = link.parent_joint
            if joint.id >= robot.dof:
                break
            jposes[:, :, joint.id : joint.id + 1] = SE3.adj(poses[link.id]) @ joint.axis

        return jposes

    class ForwardKinematics(torch.autograd.Function):
        generate_vmap_rule = True

        @staticmethod
        def forward(angles):
            poses = _forward_kinematics_helper(angles)
            rets = tuple(poses[id] for id in link_ids)
            return *rets, poses

        @staticmethod
        def setup_context(ctx, inputs, output):
            ctx.save_for_backward(*output[:-1])
            ctx.poses = output[-1]

        @staticmethod
        def backward(ctx, *grad_outputs):
            rets: tuple(torch.Tensor) = ctx.saved_tensors
            if not hasattr(ctx, "jposes"):
                ctx.jposes: torch.Tensor = _forward_kinematics_backward_helper(
                    ctx.poses
                )
            grad_pose = grad_outputs[0].new_zeros(
                grad_outputs[0].shape[0], 6, robot.dof
            )
            grad_input = grad_outputs[0].new_zeros(grad_outputs[0].shape[0], robot.dof)

            for link_id, ret, grad_output in zip(link_ids, rets, grad_outputs):
                ancestor_non_fixed_joint_ids = robot.links[
                    link_id
                ].ancestor_non_fixed_joint_ids
                temp = SE3.project(
                    torch.cat(
                        (grad_output @ ret.transpose(1, 2), grad_output[:, :, 3:]),
                        dim=-1,
                    )
                ).unsqueeze(-1)
                grad_pose[:, :, ancestor_non_fixed_joint_ids] += temp
            grad_input[:, joint_ids] = (
                ctx.jposes[:, :, joint_ids] * grad_pose[:, :, joint_ids]
            ).sum(-2)

            return grad_input

    return (
        ForwardKinematics,
        _forward_kinematics_impl,
        _forward_kinematics_helper,
        _forward_kinematics_backward_helper,
        link_ids,
        joint_ids,
        related_link_ids,
    )


def get_forward_kinematics_fns(robot: Robot, link_names: Optional[List[str]] = None):
    (
        ForwardKinematics,
        _,
        _,
        backward_helper,
        link_ids,
        _,
        related_link_ids,
    ) = ForwardKinematicsFactory(robot, link_names)

    links = robot.get_links()
    selected_link_names = [links[id].name for id in related_link_ids]
    SelectedForwardKinematics, *_ = ForwardKinematicsFactory(robot, selected_link_names)

    def forward_kinematics(angles: torch.Tensor):
        output = ForwardKinematics.apply(angles)
        return output[:-1]

    def forward_kinematics_spatial_jacobian(angles: torch.Tensor):
        selected_poses: torch.Tensor = SelectedForwardKinematics.apply(angles)
        poses = [None] * robot.num_links
        for id, pose in zip(related_link_ids, selected_poses):
            poses[id] = pose
        jposes: torch.Tensor = backward_helper(poses)

        rets = tuple(poses[id] for id in link_ids)
        jacs_s: List[torch.Tensor] = []

        for link_id in link_ids:
            pose = poses[link_id]
            jac_s = jposes.new_zeros(angles.shape[0], 6, robot.dof)
            sel = robot.links[link_id].ancestor_non_fixed_joint_ids
            jac_s[:, :, sel] = jposes[:, :, sel]
            jacs_s.append(jac_s)

        return jacs_s, rets

    def forward_kinematics_body_jacobian(angles: torch.Tensor):
        jacs_s, rets = forward_kinematics_spatial_jacobian(angles)
        jacs_b: List[torch.Tensor] = []

        for jac_s, pose in zip(jacs_s, rets):
            jacs_b.append(SE3.adj(SE3.inv(pose)) @ jac_s)

        return jacs_b, rets

    return (
        forward_kinematics,
        forward_kinematics_body_jacobian,
        forward_kinematics_spatial_jacobian,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Dict, List, Optional

import torch
import urdf_parser_py.urdf as urdf

from torchlie.functional import SE3
from torchlie.functional.constants import DeviceType

from .joint import FixedJoint, Joint, Link, PrismaticJoint, RevoluteJoint


class Robot(abc.ABC):
    def __init__(self, name: str, dtype: torch.dtype = None, device: DeviceType = None):
        self._name: str = name
        self._dtype: torch.dtype = (
            dtype if dtype is not None else torch.get_default_dtype()
        )
        self._device: torch.device = (
            torch.device(device) if device is not None else torch.device("cpu")
        )
        self._dof: int = 0
        self._num_joints: int = 0
        self._num_links: int = 0
        self._joints: List[Joint] = []
        self._links: List[Link] = []
        self._joint_map: Dict[str, Joint] = {}
        self._link_map: Dict[str, Link] = {}

    @staticmethod
    def from_urdf_file(
        urdf_file: str, dtype: torch.dtype = None, device: DeviceType = None
    ) -> "Robot":
        if dtype is None:
            dtype = torch.get_default_dtype()
        if device is None:
            device = torch.device("cpu")

        urdf_model = urdf.URDF.from_xml_file(urdf_file)
        robot = Robot(urdf_model.name, dtype, device)

        def get_joint_type(joint: urdf.Joint):
            if joint.type == "revolute" or joint.type == "continuous":
                return RevoluteJoint
            elif joint.type == "prismatic":
                return PrismaticJoint
            elif joint.type == "fixed":
                return FixedJoint
            else:
                raise ValueError(f"{joint.type} is currently not supported.")

        def get_origin(urdf_origin: Optional[urdf.Pose] = None) -> torch.Tensor:
            origin = torch.eye(3, 4, dtype=dtype, device=device).unsqueeze(0)
            if urdf_origin is None:
                return origin

            if urdf_origin.xyz is not None:
                origin[:, :, 3] = origin.new_tensor(urdf_origin.xyz)

            if urdf_origin.rpy is not None:
                rpy = origin.new_tensor(urdf_origin.rpy)
                c3, c2, c1 = rpy.cos()
                s3, s2, s1 = rpy.sin()
                origin[:, 0, 0] = c1 * c2
                origin[:, 0, 1] = (c1 * s2 * s3) - (c3 * s1)
                origin[:, 0, 2] = (s1 * s3) + (c1 * c3 * s2)
                origin[:, 1, 0] = c2 * s1
                origin[:, 1, 1] = (c1 * c3) + (s1 * s2 * s3)
                origin[:, 1, 2] = (c3 * s1 * s2) - (c1 * s3)
                origin[:, 2, 0] = -s2
                origin[:, 2, 1] = c2 * s3
                origin[:, 2, 2] = c2 * c3

            return origin

        def create_links():
            for urdf_link in urdf_model.links:
                link = Link(urdf_link.name)
                robot._link_map[urdf_link.name] = link

        def create_joints():
            for urdf_joint in urdf_model.joints:
                origin = get_origin(urdf_joint.origin)
                joint_type = get_joint_type(urdf_joint)
                parent = robot.link_map[urdf_joint.parent]
                child = robot.link_map[urdf_joint.child]
                if joint_type is FixedJoint:
                    joint = joint_type(
                        urdf_joint.name,
                        parent_link=parent,
                        child_link=child,
                        origin=origin,
                    )
                else:
                    axis = origin.new_tensor(urdf_joint.axis)
                    joint = joint_type(
                        urdf_joint.name,
                        axis,
                        parent_link=parent,
                        child_link=child,
                        origin=origin,
                    )
                child._parent_joint = joint
                parent._add_child_joint(joint)
                robot._joint_map[urdf_joint.name] = joint

        # move the child joints of any fixed joint to its parent link
        def simplify_kinematics_tree():
            for _, link in robot.link_map.items():
                for joint in link._child_joints:
                    if isinstance(joint, FixedJoint):
                        subjoints: List[Joint] = joint.child_link.child_joints
                        # removing children of fixed joint to avoid traversing twice
                        joint.child_link._child_joints = []
                        for subjoint in subjoints:
                            subjoint._parent_link = link
                            subjoint._origin = SE3.compose(
                                joint.origin, subjoint.origin
                            )
                            link._child_joints.append(subjoint)

        def update_kinematics_tree(joint: Joint):
            # update joint
            joint._id = len(robot._joints)
            robot._joints.append(joint)
            robot._dof += joint.dof
            # update link
            joint.child_link._id = len(robot._links)
            robot._links.append(joint.child_link)

        def build_kinematics_tree():
            # link and joint ids are assigned through BFS where fixed joints are skipped
            # fixed joint ids are always greater than non-fixed joint ids
            robot._links = []
            robot._joints = []

            joints_to_visit: List[Joint] = []
            root = robot.link_map[urdf_model.get_root()]
            root._id = 0
            robot._links.append(root)
            robot._dof = 0
            joints_to_visit = joints_to_visit + root.child_joints

            while joints_to_visit:
                joint = joints_to_visit.pop(0)
                # skip fixed joint when assigning joint id
                if not isinstance(joint, FixedJoint):
                    # update degrees of freedom
                    update_kinematics_tree(joint)
                    joints_to_visit = joints_to_visit + joint.child_link.child_joints

            for _, joint in robot.joint_map.items():
                if joint.id is not None:
                    continue
                # non-fixed joints should have already been processed
                if not isinstance(joint, FixedJoint):
                    raise ValueError(f"{joint.name} is expected to a fixed joint.")
                # fixed joint ids are greater than these of non-fixed ones
                update_kinematics_tree(joint)

        # cache ancestor non-fixed joints to ease FK computation
        def cache_ancestor_non_fixed_joint_ids():
            for link in robot.links:
                if link.parent_joint is not None:
                    link._ancestor_links = link.parent_link.ancestor_links + [
                        link.parent_link
                    ]
                    ancestor_non_fixed_joint_ids = (
                        link.parent_link.ancestor_non_fixed_joint_ids
                        if isinstance(link.parent_joint, FixedJoint)
                        else link.parent_link.ancestor_non_fixed_joint_ids
                        + [link.parent_joint.id]
                    )
                    link._ancestor_non_fixed_joint_ids = ancestor_non_fixed_joint_ids

        create_links()
        create_joints()
        simplify_kinematics_tree()
        build_kinematics_tree()

        robot._num_links = len(robot.links)
        robot._num_joints = len(robot.joints)

        cache_ancestor_non_fixed_joint_ids()

        return robot

    def get_links(self, link_names: Optional[List[str]] = None) -> List[Link]:
        return (
            self.links
            if link_names is None
            else [self.link_map[name] for name in link_names]
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def dtype(self) -> torch.dtype:
        return self._dtype

    @property
    def device(self) -> torch.device:
        return self._device

    @property
    def dof(self) -> int:
        return self._dof

    @property
    def num_joints(self) -> int:
        return self._num_joints

    @property
    def num_links(self) -> int:
        return self._num_links

    @property
    def joints(self) -> List[Joint]:
        return self._joints

    @property
    def links(self) -> List[Link]:
        return self._links

    @property
    def joint_map(self) -> Dict[str, Joint]:
        return self._joint_map

    @property
    def link_map(self) -> Dict[str, Link]:
        return self._link_map



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import setuptools
from pathlib import Path

torchkin_path = Path("torchkin")
with open(torchkin_path / "__init__.py", "r") as f:
    for line in f:
        if "__version__" in line:
            version = line.split("__version__ = ")[1].rstrip().strip('"')

with open("README.md", "r") as fh:
    long_description = fh.read()


def parse_requirements_file(path):
    with open(path) as f:
        reqs = [line.strip() for line in f]
    return reqs


reqs_main = parse_requirements_file("requirements.txt")
setuptools.setup(
    name="torchkin",
    version=version,
    author="Meta Research",
    description="Torch extension for differentiable kinematics.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/facebookresearch/theseus/lie",
    keywords="lie groups, differentiable optimization",
    packages=["torchkin"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    install_requires=reqs_main,
    python_requires=">=3.8",
)
